// prefix header, with public (user-code-reachable) struct and typedef declarations
#include "FBtoC.h"
#include "Runtime.h"

#if !DECARBONATE

/*
 FileHandling.c
 */


#define FBERRORS 13

OSStatus			gFBFileError;
OSType				gFBSaveType    = 'TEXT';						// used by FBSetOpenType and other 
OSType				gFBSaveCreator = 'ttxt';						// used by FBSetOpenType and other 
FSRef				gFBDefaultFolderRef; 
FSSpec				gFBDefaultFolderSpec; 

Str255				gFBLineBuf;
UInt32				gFBLineDev;
UInt32				gFBLinePos;
SInt64              gFBLastFilePos;
char                gFBFileTabMode;
UInt16              gFBTabWidth = 16;

#if NO_SUFFIX_ON_ARRAY_NAMES
FBInfoForOpenFile	gFBOpenFiles[kMaxFBFiles + kFBMaxSerialFiles]; // suffix matches FB's dim as FBInfoForOpenFile	  gFBOpenFiles(_kFBMaxDiskFiles + _kFBMaxSerialFiles) 
#else
FBInfoForOpenFile	gFBOpenFiles_A[kMaxFBFiles + kFBMaxSerialFiles]; // suffix matches FB's dim as FBInfoForOpenFile	  gFBOpenFiles(_kFBMaxDiskFiles + _kFBMaxSerialFiles) 
#endif

static OSStatus		FBReadFile_Priv( const int fileID, void * dest, ByteCount numBytesRequested, ByteCount * numBytesRead );
static int			FBXSerialRead_Priv( const int serialPortID );
static void			FBCloseOneFile_priv( int fileID );


/*
 Set gFBDefaultFolderRef to the FSRef for folder enclosing the application
 */

void FBInitDefaultFolderRef( void  )
{
	FSRef	appRef;
	ProcessSerialNumber	psn = { 0, kCurrentProcess };
	GetProcessBundleLocation( &psn, &appRef );
	FSGetCatalogInfo( &appRef, kFSCatInfoNone, NULL, NULL, &gFBDefaultFolderSpec, &gFBDefaultFolderRef );
}


////////////////////////////////////
//      !!!: Error handling        //
////////////////////////////////////



OSStatus FBerror( void )
{
	/*
	 error function
	 */
	return gFBFileError;
}




void FBFileErrorAlert( int fileID, OSStatus err )
{
	Str255	msg;
	
	if ( err )
	{		
		PSstrcpy( TO_STACK,  PSltrr( "The program has returned a file error.", "FBFileErrorAlert" ) ), PSstrcat( TO_STACK, PSstring( 1, 13 ) ), PSstrcat( TO_STACK, PSstr( err ) ), PSstrcat( TO_STACK, PSstring( 1, 13 ) ); 
		PSstrcpy( msg, TO_STACK );
		PSstrcpy( TO_STACK, msg ), PSstrcat( TO_STACK, PSltrr( "File ref:", "FBFileErrorAlert" ) ), PSstrcat( TO_STACK, PSstr( fileID ) ), PSstrcat( TO_STACK, PSstring( 1, 13 ) ); 
		PSstrcpy( msg, TO_STACK );
		FBStopMsg( PSstrcpy( TO_STACK, msg ), 0 );
		
	}
}


/*
 value may be: FBFileErrorAlert (the default), a user-function (from 'on error fn MyErrHandler'), or NULL (from 'on error end').
 */
FBOnXxxxxVector		gFBOnErrorVector = &FBFileErrorAlert; // [fn name marks line to be removed by C runtime stripping] FBFileErrorAlert()



void FBCheckFileError( int fileID, OSStatus err )
{
	/*
	 all opens, reads, writes   call though here at exit
	 
	 if ( err == noErr ) do nothing
	 if ( gFBOnErrorVector is 0 ) do nothing  else call the vector
	 */
	if ( err )
	{
		gFBFileError = err;
		if ( gFBOnErrorVector ) (*gFBOnErrorVector)( fileID, err ); 
	}
}


////////////////////////////////////
//           !!!: Open / close    //
////////////////////////////////////


void FBOpenUNIX( unsigned fileID, Str255 command )
{
	/*
	 Open "UNIX", fileID, command
	 
	 */
	gFBStk--;
	Str255			eStr1,eStr2;
	OSStatus		err = noErr;
	
	if ( fileID > 0 && fileID < kMaxFBFiles )
	{
		if ( gFBOpenFiles_A[fileID].isOpen == false )
		{
			PScstr( command );
			gFBOpenFiles_A[fileID].pipeUNIX = popen( (char*)command, "r" );
			gFBOpenFiles_A[fileID].isOpen = true;
		}
		else
		{
			PSstrcpy( eStr1, PSltrr( "Error in Open 'UNIX' for file# ", "FBOpenUNIX" ) );
			PSstrcat( eStr1, PSstr( fileID ) );
			PSstrcpy( eStr2, PSltrr( "FileID already in use", "FBOpenUNIX") );
			FBfullStop( eStr1, eStr2, false );
		}
		
	}
	else
	{
		PSstrcpy( eStr1, PSltrr( "FileID not in range ", "FBOpenUNIX" ) );
		PSstrcat( eStr1, PSstr( err ) );
		PSstrcat( eStr1, PSltrr( " for FileID: ", "FBOpenUNIX" ) );
		PSstrcat( eStr1, PSstr( fileID ) );
		FBfullStop( eStr1, eStr2, false );
		
	}
}




void FBOpenOutput( const int fileID, void *fileNamePtr, void *fs )
{
	/*
	 open "O", 2, filename, @fspec  open "O", #2, filename, @fspec    open "O", varname, filename, @fspec     formats
	 Works for data fork only
	 Bad fileID gives no errors, just does nothing harmlessly.
	 Other errors give no message, but plough on regardless
	 
	 Optional:
	 FSRef: open "O", fileID, fileName, @ref
	 CFURLRef: open "O", fileID, @url
	 */
	FSRef			parRef, fileRef;
#if !__LP64__
	FSSpec			parSpec, spec;
#endif  /* !__LP64__ */
	CFURLRef		url, parURL;
	OSStatus		err = noErr;
    CFIndex			len;
	CFStringRef		saveFileName;
	FSCatalogInfo	catInfo;
	FileInfo		*finderInfo;
	HFSUniStr255	dataForkName;
	UniChar         fileName[255];
#if MAC_OS_X_VERSION_10_4 < MAC_OS_X_VERSION_MAX_ALLOWED
	FSIORefNum		forkRefNum;
#else
	short			forkRefNum;
#endif
	Str255			eStr1,eStr2;
	
	if ( fileID > 0 && fileID < kMaxFBFiles )
	{
		switch ( (long)fileNamePtr )
		{
#if !__LP64__
			case 0:																// FSSpec
				BlockMoveData( fs, &spec, sizeof(FSSpec) );
				
				// get dir from FSSpec and get FSRef for dir 
				err			 = FSMakeFSSpec( spec.vRefNum, spec.parID, "\p", &parSpec ); 
				require_noerr( err, Bail );
				err			 = FSpMakeFSRef( &parSpec, &parRef );
				require_noerr( err, Bail );
				
				// create CFString from FSSpec name	     				  
				saveFileName = CFStringCreateWithPascalString( kCFAllocatorDefault, spec.name, kCFStringEncodingMacRoman );
				len			 = CFStringGetLength( saveFileName );
				CFStringGetCharacters( saveFileName, CFRangeMake(0, len), fileName );
				CFRelease( saveFileName );
				
				// FSRef for the file we want to create
				err = FSMakeFSRefUnicode( &parRef, len, fileName, kTextEncodingUnknown, &fileRef );			
				
				// create new file - fileRef is new file.
				if ( err == fnfErr ) err = FSCreateFileUnicode( &parRef, len, fileName, kFSCatInfoNone, NULL, &fileRef, NULL );				
				break;
#endif  /* !__LP64__ */
				
			case 2:																// CFURLRef
				BlockMoveData( fs, &url, sizeof(CFURLRef) );
				if ( CFURLGetFSRef( url, &fileRef ) == false ) // file does not exist; create it
				{
					saveFileName = CFURLCopyLastPathComponent( url );
					len = CFStringGetLength( saveFileName );
					CFStringGetCharacters( saveFileName, CFRangeMake( 0, len ), fileName );
					CFRelease( saveFileName );
					
					parURL = CFURLCreateCopyDeletingLastPathComponent( kCFAllocatorDefault, url );
					if ( parURL )
					{
						CFURLGetFSRef( parURL, &parRef );
						err = FSMakeFSRefUnicode( &parRef, len, fileName, kTextEncodingUnknown, &fileRef );
						if ( err == fnfErr )					// file does not exist; create it
						{
							err = FSCreateFileUnicode( &parRef, len, fileName, kFSCatInfoNone, NULL, &fileRef, NULL );
						}
						CFRelease( parURL );
					}
				}
				break;
				
			default:															// FSRef
				BlockMoveData( fs, &parRef, sizeof(FSRef) );
				saveFileName = CFStringCreateWithPascalString( kCFAllocatorDefault, fileNamePtr, kCFStringEncodingMacRoman );
				gFBStk -= 1; // was incremented by caller
				len = CFStringGetLength( saveFileName );
				CFStringGetCharacters( saveFileName, CFRangeMake( 0, len ), fileName );
				CFRelease( saveFileName );
				
				err = FSMakeFSRefUnicode( &parRef, len, fileName, kTextEncodingUnknown, &fileRef );
				if ( err == fnfErr ) 					// file does not exist; create it
					
				{
					err = FSCreateFileUnicode( &parRef, len, fileName, kFSCatInfoNone, NULL, &fileRef, NULL );
				}
				break;
		}
		
		require_noerr( err, Bail );
		
		// Get the FinderInfo data, set based on DEF OPEN globals
		(void)FSGetCatalogInfo( &fileRef, kFSCatInfoFinderInfo, &catInfo, NULL, NULL, NULL );         // get finder info
		
		finderInfo	 = (FileInfo*) catInfo.finderInfo;
		finderInfo->fileType    = gFBSaveType;            
		finderInfo->fileCreator = gFBSaveCreator;         
		err			 = FSSetCatalogInfo( &fileRef, kFSCatInfoFinderInfo, &catInfo );  	
		require_noerr( err, Bail );
		
		err			 = FSGetDataForkName( &dataForkName );
		require_noerr( err, Bail );
		
		err			 = FSOpenFork( &fileRef, dataForkName.length, dataForkName.unicode, fsRdWrPerm, &forkRefNum );
		require_noerr( err, Bail );
		
		FSSetForkSize( forkRefNum, fsFromStart, 0 );
		// Stash for future use
		gFBOpenFiles_A[fileID].isOpen		= true;
		gFBOpenFiles_A[fileID].fileRef		= fileRef;
		gFBOpenFiles_A[fileID].forkRefNum	= forkRefNum;
		gFBOpenFiles_A[fileID].recordSize	= 256;     // default until fn OpenOutputSubStatement enhanced
		gFBOpenFiles_A[fileID].linePos   	= 0;  //20080131
	}
	else
	{
		PSstrcpy( eStr1, PSltrr( "Error in FBOpenOutput for file# ", "FBOpenOutput" ) );
		PSstrcat( eStr1, PSstr( fileID ) );
		PSstrcpy( eStr2, PSltrr( "FileID must be in range 1 to ", "FBOpenOutput" ) );
		PSstrcat( eStr2, PSstr( kMaxFBFiles - 1 ) );
		FBfullStop( eStr1, eStr2, false );
	}
Bail:
	FBCheckFileError( fileID, err );
}



void FBOpenRandom( const int fileID, void *fileNamePtr, void *fs, SInt64 recLen, Boolean exclusive )
{
	/*
	 1) Open "R",2,@fs,reclen  or    Open "R",#2,@fs,reclen  or     Open "R",varName,@fs,reclen formats only
	 Open "N" formats added rp:20091006 
	 
	 Brian 20091201 -  bug #407 If the user data file does not exist and the user specified the FSSpec or CFURLRef version, the file is created. If the user specifies an
	 FSRef for a non existent file, the runtime catches it with an error
	 */
	FSRef			fRef;
#if MAC_OS_X_VERSION_10_4 < MAC_OS_X_VERSION_MAX_ALLOWED
	FSIORefNum		forkRefNum;
#else
	short			forkRefNum;
#endif
	HFSUniStr255	dataForkName;
	OSStatus		err = noErr;
#if !__LP64__
	FSSpec			spec;
#endif  /* !__LP64__ */
	CFURLRef		url;
	SInt8			permissions = fsRdWrShPerm;
	
	if ( fileID > 0 && fileID < kMaxFBFiles )
	{
		switch ( (long)fileNamePtr )
		{
#if !__LP64__
			case 0:																// FSSpec
				BlockMoveData( fs, &spec, sizeof(FSSpec) );
				err = FSpMakeFSRef( &spec, &fRef );
				if ( err == fnfErr )
				{
					FBOpenOutput( fileID, fileNamePtr, fs );
					FBClose( fileID );
					err = FSpMakeFSRef( &spec, &fRef );
				}
				
				require_noerr( err, Bail );				
				break;
#endif  /* !__LP64__ */
				
			case 1:																// FSRef
				BlockMoveData( fs, &fRef, sizeof(FSRef) );
				break;
				
			case 2:																// CFURLRef
				BlockMoveData( fs, &url, sizeof(CFURLRef) );
				if ( CFURLGetFSRef( url, &fRef ) == false )
				{
					FBOpenOutput( fileID, fileNamePtr, fs );
					FBClose( fileID );
					CFURLGetFSRef( url, &fRef );
				}
				break;			
		}
		
		err = FSGetDataForkName( &dataForkName );
		require_noerr( err, Bail );
		
		if ( exclusive ) permissions = fsRdWrPerm;
		err = FSOpenFork( &fRef, dataForkName.length, dataForkName.unicode, permissions, &forkRefNum );
		require_noerr( err, Bail );
		
		gFBOpenFiles_A[fileID].isOpen		= true;
		gFBOpenFiles_A[fileID].fileRef		= fRef;
		gFBOpenFiles_A[fileID].forkRefNum	= forkRefNum;
		gFBOpenFiles_A[fileID].recordSize	= recLen;  
		gFBOpenFiles_A[fileID].linePos   	= 0;  //20080131
	}
Bail:
	FBCheckFileError( fileID, err );
}


void FBOpenAppend( const int fileID, void *fileNamePtr, void *fs, SInt64 recLen )
{
	/*
	 1) Open "A",2,@fs
	 2) Assumes data fork (i.e.  "I") and does not accept "ID" or "IR" for data/resource fork
	 Brian 1-August-2007
	 
	 mods for open "A", fileID, @ref and open "A", fileID, @url  Brian 20080917
	 
	 Brian 20091129 -  bug #376 If the user data file does not exist and the user specified the FSSpec or CFURLRef version, the file is created. If the user specifies an
	 FSRef for a non existent file, the runtime catches it with an error
	 */
	OSStatus		err;
#if !__LP64__
	FSSpec			spec;
#endif  /* !__LP64__ */
	CFURLRef		url;
	FSRef			fRef;
	
	switch ( (long)fileNamePtr ) {
#if !__LP64__
		case 0:																// FSSpec
			BlockMoveData( fs, &spec, sizeof(FSSpec) );
			err = FSpMakeFSRef( &spec, &fRef );
			if ( err == fnfErr )
			{
				FBOpenOutput( fileID, fileNamePtr, fs );
				FBClose( fileID );
				err = noErr;
			}
			require_noerr( err, Bail );				
			break;
#endif  /* !__LP64__ */
		case 2:																// CFURLRef
			BlockMoveData( fs, &url, sizeof(CFURLRef) );
			if ( CFURLGetFSRef( url, &fRef ) == false )
			{
				FBOpenOutput( fileID, fileNamePtr, fs );
				FBClose( fileID );
			}
			break;
			
		default:
			break;
	}
	
	FBOpenRandom( fileID, fileNamePtr, fs, recLen, true );
	err = FSSetForkPosition( gFBOpenFiles_A[fileID].forkRefNum, fsFromLEOF, 0 );
Bail:
	FBCheckFileError( fileID, err );
}


void FBOpenInput( const int fileID, void *fileNamePtr, void *fs, SInt64 recLen, Boolean isResourceOpen  )
{
	/*
	 Open "I",2,@fs  or    Open "I",#2,@fs  or     Open "I",varName,@fs formats only
	 
	 Optional (FBtoC only)
	 FSRef:  open "I", fileID, @ref
	 CFURLRef: open "I", fileID, @url
	 */
	FSRef			fRef;
#if !__LP64__
	FSSpec			spec;
#endif  /* !__LP64__ */
	CFURLRef		url;
#if MAC_OS_X_VERSION_10_4 < MAC_OS_X_VERSION_MAX_ALLOWED
	FSIORefNum		forkRefNum;
#else
	short			forkRefNum;
#endif
	HFSUniStr255	forkName;
	OSStatus		err = noErr;
	
	if ( fileID > 0 && fileID < kMaxFBFiles )
	{
		
		switch ( (long)fileNamePtr )
		{
#if !__LP64__
			case 0:																// FSSpec
				BlockMoveData( fs, &spec, sizeof(FSSpec) );
				err = FSpMakeFSRef( &spec, &fRef );
				if ( err == fnfErr ) err = noErr;
				require_noerr( err, Bail );				
				break;
#endif  /* !__LP64__ */
			case 1:																// FSRef
				BlockMoveData( fs, &fRef, sizeof(FSRef) );
				break;
				
			case 2:																// CFURLRef
				BlockMoveData( fs, &url, sizeof(CFURLRef) );
				CFURLGetFSRef( url, &fRef );
				break;			
		}
		
		if ( isResourceOpen )
			err = FSGetResourceForkName( &forkName );
		else
			err = FSGetDataForkName( &forkName );
		require_noerr( err, Bail );
		
		err = FSOpenFork( &fRef, forkName.length, forkName.unicode, fsRdPerm, &forkRefNum );  // Open the fork read only
		require_noerr( err, Bail );
		
		gFBOpenFiles_A[fileID].isOpen		= true;
		gFBOpenFiles_A[fileID].fileRef		= fRef;
		gFBOpenFiles_A[fileID].forkRefNum	= forkRefNum;
		gFBOpenFiles_A[fileID].recordSize	= recLen;  
		gFBOpenFiles_A[fileID].linePos   	= 0; 
	}
Bail:
	FBCheckFileError( fileID, err );
}




static void FBCloseOneFile_priv( int fileID )
{    
    if ( gFBOpenFiles_A[fileID].pipeUNIX ) // UNIX
    {
        pclose( gFBOpenFiles_A[fileID].pipeUNIX );
    }
    else if ( gFBOpenFiles_A[fileID].forkRefNum ) // file
    {
        gFBLinePos = 0;
        FSCloseFork( gFBOpenFiles_A[fileID].forkRefNum );
    }
    else if ( gFBOpenFiles_A[fileID].fileDescriptor ) // serial
    {
        gFBLinePos = 0;
        close( gFBOpenFiles_A[fileID].fileDescriptor );
        free( gFBOpenFiles_A[fileID].buff );
    }
    BlockZero( &gFBOpenFiles_A[fileID], sizeof( FBInfoForOpenFile ) );
}


void FBClose( int fileID )
{	
	/*
	 close fileID
	 */
	if ( fileID == 0 )	// close all
	{
		for ( fileID = 1; fileID < (kMaxFBFiles + kFBMaxSerialFiles); fileID++ ) { FBCloseOneFile_priv ( fileID ); }
	}
	else if ( fileID >= -kFBMaxSerialFiles && fileID < kMaxFBFiles )
	{
		if ( fileID < 0 ) fileID = kMaxFBFiles - fileID;
		FBCloseOneFile_priv( fileID );
	}
	
}



////////////////////////////////////
//     !!!: Enquiries / commands  //
////////////////////////////////////


long FBeof( int fileID )
{
	FBBoolean		result = ZTRUE;
	
	if ( fileID > 0 && fileID < kMaxFBFiles && gFBOpenFiles_A[fileID].isOpen )
	{
		if ( gFBOpenFiles_A[fileID].pipeUNIX ) // UNIX
		{
			result = gFBOpenFiles_A[fileID].eof;
		}
		else								// file
		{
			SInt64		size, forkPos;
			
			if ( noErr != FSGetForkPosition( gFBOpenFiles_A[fileID].forkRefNum, &forkPos) ) { forkPos = 0; }
			if ( noErr != FSGetForkSize( gFBOpenFiles_A[fileID].forkRefNum, &size ) ) { size = 0; }
			//printf( "FBeof  fileID %d  size  %d, forkPos %d\n", fileID, (int)size, (int)forkPos );
			if ( forkPos < size ) result = false;
		}
	}
	return result;
}



SInt64 FBlof( int fileID, SInt64 recLen )
{
	SInt64			size = 0;
	int				serialPortID;
	
	if ( fileID > 0 && fileID < kMaxFBFiles && (gFBOpenFiles_A[fileID].forkRefNum) )
	{
		if ( noErr != FSGetForkSize( gFBOpenFiles_A[fileID].forkRefNum, &size ) ) { size = 0; }
	}
	
	else if ( fileID < 0 && fileID >= -kFBMaxSerialFiles ) // serial
	{
		serialPortID = kMaxFBFiles - fileID;
		FBXSerialRead_Priv( serialPortID ); // grab any chars pending
		size = gFBOpenFiles_A[serialPortID].bytesInBuff;
	}
	return size/recLen;  
}


void FBAppend( const int fileID )
{
	// FB append statement
	OSStatus		err;
	
	err = FSSetForkPosition( gFBOpenFiles_A[fileID].forkRefNum, fsFromLEOF, 0 );
	FBCheckFileError( fileID, err );
}


void FBRecord( const int fileID, SInt64 recNum, SInt64 posInRec )
{
	/*
	 Record [#]fileID, recordNum [,positionInRecord]
	 */
	SInt64			newPos;
	OSStatus		err  = noErr;
	
	if ( fileID > 0 && fileID < kMaxFBFiles && (gFBOpenFiles_A[fileID].forkRefNum) )
	{
		newPos = recNum * gFBOpenFiles_A[fileID].recordSize + posInRec;
		err = FSSetForkPosition( gFBOpenFiles_A[fileID].forkRefNum, fsFromStart, newPos );
	}
	if ( err == eofErr ) // extend the file
	{
		FSForkIOParam    pb;
		BlockZero( &pb, sizeof( pb ) );
		pb.forkRefNum     = gFBOpenFiles_A[fileID].forkRefNum;
		pb.positionMode   = fsFromStart;
		pb.positionOffset = newPos;
		err = PBSetForkSizeSync( &pb );
		if ( err == noErr )
        {
            err = FSSetForkPosition( gFBOpenFiles_A[fileID].forkRefNum, fsFromStart, newPos );    //Paolo 4-02-09
        }		
	}
	FBCheckFileError( fileID, err );		
}


SInt64 FBrec( int fileID )
{
	// FB rec statement
	SInt64			forkPos = 0;
	
	if ( fileID > 0 && fileID < kMaxFBFiles && gFBOpenFiles_A[fileID].isOpen )
	{
		if ( noErr != FSGetForkPosition( gFBOpenFiles_A[fileID].forkRefNum, &forkPos) ) { forkPos = 0; }
	}
	return forkPos/gFBOpenFiles_A[fileID].recordSize;
}


SInt64 FBloc( int fileID )
{
	// FB loc function
	SInt64			forkPos = 0;
	
	if ( fileID > 0 && fileID < kMaxFBFiles && gFBOpenFiles_A[fileID].isOpen )
	{
		if ( noErr != FSGetForkPosition( gFBOpenFiles_A[fileID].forkRefNum, &forkPos) ) { forkPos = 0; }
	}
	return forkPos % gFBOpenFiles_A[fileID].recordSize;
}



////////////////////////////////////
//            !!!: Input          //
////////////////////////////////////


void FBLineInputHash( int fileID, StringPtr s )
{
	/*
	 line input #fileID, stringVar$
	 */
	long			len;
	signed char		theChar;
	Boolean			eol;
	OSStatus        err = noErr;
	
	*(char*)s = 0; // null string if error
	
	//	if ( fileID == 0 )
	//	{
	//		//FBLineInput( "\p", s ); // antique Terminal.app-hosted stdin stdout thing
	//	}
	//	else
	if ( fileID < 0 ) // serial
	{
		do
		{
			int bytesRead = FBReadFile( fileID, &theChar, false, 1 );
			if ( bytesRead == 0 ) break;
			if ( theChar != 13  )
			{
				s[0]++;
				s[s[0]] = theChar;
			}
		}
		while ( theChar != 13 && s[0] < 255 );       

	}
	else	
	{
		if ( fileID > 0 && fileID < kMaxFBFiles ) // UNIX or file
		{
			if (gFBOpenFiles_A[fileID].pipeUNIX) // UNIX
			{
				eol = false;
				len = 0;
				do
				{
					theChar = fgetc( gFBOpenFiles_A[fileID].pipeUNIX );
					switch ( theChar ) 
					{
						case 10:
						case 13:
							eol = true;
							break;
						case EOF:
							eol = true;
							gFBOpenFiles_A[fileID].eof = ZTRUE;
							break;
						default:
							len++;
							s[len] = theChar;							
					}
				}
				while ( eol == false && len < 255 );
				s[0] = len;
			}
			else // file
			{
				if ( gFBOpenFiles_A[fileID].forkRefNum )
				{
					
					unsigned char	ch;
					int				count = 0;
					
					do
					{
						err = FSReadFork( gFBOpenFiles_A[fileID].forkRefNum, fsAtMark + pleaseCacheMask, 0, 1, &ch, NULL );
						if ( err ) break;
						count++;
						*(char*)(s + count) = ch;
#define LINE_INPUT_TERMINATED_BY_LINEFEED 1
#if LINE_INPUT_TERMINATED_BY_LINEFEED
					} while ( count < 255 && ch != 13 && ch != 10 );
					if ( !err && (ch == 13 || ch == 10) ) { count--; }			// strip <cr> or <lf>
#else
					} while ( count < 255 && ch != 13 );
					if ( !err && ch == 13 ) { count--; }			// strip <cr>
#endif
					*(char*)s = count;								// length byte
					if ( count && (err == eofErr) ) err = noErr;	// if we read any char at all, don't signal eof error
				}
				else
				{
					err = fnOpnErr; //  file not open 
				}
			}
		}
	}
	FBCheckFileError( fileID, err );
}




static OSStatus FBReadFile_Priv( const int fileID, void * dest, ByteCount numBytesRequested, ByteCount * numBytesRead )
{	
	OSStatus		err = noErr;
	int				serialPortID;
	
	*numBytesRead = 0; // default if error
	if ( fileID > 0 && fileID < kMaxFBFiles ) // disk file
	{
		if ( gFBOpenFiles_A[fileID].isOpen && gFBOpenFiles_A[fileID].forkRefNum )
		{
			err = FSReadFork( gFBOpenFiles_A[fileID].forkRefNum, fsAtMark + pleaseCacheMask, 0, numBytesRequested, dest, numBytesRead );
		}
		else
		{ err = fnOpnErr; }
	}
	
	else if ( fileID < 0 && fileID >= -kFBMaxSerialFiles ) // serial
	{
		serialPortID = kMaxFBFiles - fileID;
		if ( gFBOpenFiles_A[serialPortID].isOpen )
		{
			FBXSerialRead_Priv( serialPortID );
			if ( gFBOpenFiles_A[serialPortID].bytesInBuff >= numBytesRequested )
			{
				BlockMoveData( gFBOpenFiles_A[serialPortID].buff, dest, numBytesRequested ); // copy from serial buffer to dest
				gFBOpenFiles_A[serialPortID].bytesInBuff -= numBytesRequested;
				BlockMoveData( gFBOpenFiles_A[serialPortID].buff + numBytesRequested, gFBOpenFiles_A[serialPortID].buff, gFBOpenFiles_A[serialPortID].bytesInBuff ); // 'left-justify' buffer contents
				*numBytesRead = numBytesRequested;
			}
			else
			{ err = eofErr; } // asked for too many
		}
		else
		{ err = fnOpnErr; }
	}
	return err;
}



ByteCount FBReadFile( const int fileID, void * dest, Boolean isInputHashString, ByteCount numBytesRequested )
{
	OSStatus	err;
	ByteCount	numBytesRead;
	
	if ( isInputHashString && (numBytesRequested == 0) ) numBytesRequested = 1; // read #serialPort, s$; 0   behaves like  inkey$( serialPort )
	err = FBReadFile_Priv( fileID, dest, numBytesRequested, &numBytesRead );
	if ( isInputHashString == false ) FBCheckFileError( fileID, err );
	return numBytesRead;
}


StringPtr FBInkeyFromFile( const int fileID )
{
	// inkey$() return 1 character on success, null string if failure; failure is silent
	ByteCount	numBytesRead;
	
	FBReadFile_Priv( fileID, &gReturnedString[1], 1, &numBytesRead );
	gReturnedString[0] = numBytesRead;
	return gReturnedString;
}

/*
 read# numeric var    big-endian on disk
 */
short FBReadSwapShort( const unsigned fileID )
{
	short s;
	FBReadFile( fileID, &s, false, 2 );
	return CFSwapInt16BigToHost( s );
}

long FBReadSwapLong( const unsigned fileID )
{
	long l;
	FBReadFile( fileID, &l, false, 4 );
	return CFSwapInt32BigToHost( l );
}

float FBReadSwapSingle( const unsigned fileID )
{
	CFSwappedFloat32 fs;
	FBReadFile( fileID, &fs, false, sizeof( fs ) );
	return CFConvertFloatSwappedToHost( fs );
}

double FBReadSwapDouble( const unsigned fileID )
{
	CFSwappedFloat64 ds;
	FBReadFile( fileID, &ds, false, sizeof( ds ) );
	return CFConvertDoubleSwappedToHost( ds );
}


void FBReadField( const int fileID, Handle * outHandle )
{
	// read field [#]deviceID, handleVar
	long              size;
	Handle            h;
	
	size = FBReadSwapLong( fileID );
	h = NewHandle( size );
	if ( h ) FBReadFile( fileID, *h, false, size );
	*outHandle = h;
} 



static SInt64 FBStringToSInt64( ConstStr255Param s )
{
	 // fixes bug #358
	return (SInt64)PSval( PSstrcpy( STACK_PUSH, s ) );
}


void FBParseInput( const short  varType, void * vPtr )
{
	void *            textPtr;
	long              linePos;
	Str255            parsedStr;
	
	textPtr = &gFBLineBuf;
	linePos = gFBLinePos;
	
	PSstrcpy( parsedStr, "\p" );
	while ( (-(linePos <= *(unsigned char*)(textPtr) ) & ( ((-(*(unsigned char*)(textPtr + linePos) == (unsigned char)' ' ) | ( -(*(unsigned char*)(textPtr + linePos) == 9 ) ) )) ) ) )
	{
		linePos++;
	}
	if ( -(*(unsigned char*)(textPtr + linePos) == PSasc( PSstrcpy( STACK_PUSH, "\p\"" ) ) ) )
	{
		linePos++;
		while ( (-(linePos <= *(unsigned char*)(textPtr) ) & ( -(*(unsigned char*)(textPtr + linePos) != PSasc( PSstrcpy( STACK_PUSH, "\p\"" ) ) ) ) ) )
		{
			(PSstrcpy( TO_STACK, parsedStr ), PSstrcat( TO_STACK, PSstring( 1, *(unsigned char*)(textPtr + linePos) ) )); PSstrcpy( parsedStr, TO_STACK );
			linePos++;
		}
		if ( (-(linePos <= *(unsigned char*)(textPtr) ) & ( -(*(unsigned char*)(textPtr + linePos) == PSasc( PSstrcpy( STACK_PUSH, "\p\"" ) ) ) ) ) )
		{
			linePos++;
		}
	}
	else
	{
		while ( (-(linePos <= *(unsigned char*)(textPtr) ) & ( (-(*(unsigned char*)(textPtr + linePos) != (unsigned char)',' ) & ( -(*(unsigned char*)(textPtr + linePos) != 9 ) ) ) ) ) )
		{
			(PSstrcpy( TO_STACK, parsedStr ), PSstrcat( TO_STACK, PSstring( 1, *(unsigned char*)(textPtr + linePos) ) )); PSstrcpy( parsedStr, TO_STACK );
			linePos++;
		}
	}
	while ( (-(linePos <= *(unsigned char*)(textPtr) ) & ( ((-(*(unsigned char*)(textPtr + linePos) == (unsigned char)' ' ) | ( (-(*(unsigned char*)(textPtr + linePos) == 9 ) | ( -(*(unsigned char*)(textPtr + linePos) == 13 ) ) ) ) )) ) ) )
	{
		linePos++;
	}
	if ( (-(linePos <= *(unsigned char*)(textPtr) ) & ( -(*(unsigned char*)(textPtr + linePos) == (unsigned char)',' ) ) ) )
	{
		linePos++;
	}
	
	// magic numbers here are varTypes coded in InitCTypetable
	
	if ( varType == 0 || varType == 1 || varType == 2 ) // _longVarType, _fixedVarType, _unsignedLongVarType
	{
		*(SInt32*)vPtr = (SInt32)FBStringToSInt64( parsedStr );
	}
	else if ( varType == 3 || varType == 4 ) // _shortVarType, _unsignedShortVarType
	{
		*(SInt16*)vPtr = (SInt16)FBStringToSInt64( parsedStr );
	}
	else if ( varType == 5 || varType == 6 || varType == 7 ) // _byteVarType, _unsignedByteVarType, _booleanVarType
	{
		*(SInt8*)vPtr = (SInt8)FBStringToSInt64( parsedStr );
	}
	else if ( varType == 8 || varType == 9 ) // _UInt64VarType, _SInt64VarType
	{
		*(SInt64*)vPtr = FBStringToSInt64( parsedStr );
	}
	// 10, 11 are _pointerVarType, _handleVarType
	else if ( varType == 12 || varType == 13 ) //_floatVarType, _singleVarType
	{
		*(float*)vPtr = PSval( PSstrcpy( STACK_PUSH, parsedStr ) );
	}
	else if ( varType == 14 ) // _doubleVarType
	{
		*(double*)vPtr = PSval( PSstrcpy( STACK_PUSH, parsedStr ) );
	}
	else if ( varType == 15 ) // _ldoubleVarType
	{
		*(long double*)vPtr = PSval( PSstrcpy( STACK_PUSH, parsedStr ) );
	}
	// 16 is _doubleComplexType
	// 17 is _constStr255VarType
	else if ( varType >= 18 && varType <= 145 ) // _str1VarType to _str255VarType
	{
		PSstrcpy( (StringPtr)vPtr, parsedStr );
	}
	else if ( varType == 146 ) // _containerVarType
	{
		CtnrStoreString( (Handle*)vPtr, parsedStr );
	}
	
	gFBLinePos = linePos;
} 


void FBToCInputFile( const int fileID, void *target, short varType )
{
	UInt32 lastDevice;
	lastDevice = gFBLineDev;
	gFBLineDev = fileID;
	
	if ( fileID )
	{
		if ( lastDevice != gFBLineDev || gFBLinePos == 0 || gFBLinePos > gFBLineBuf[0] )
		{
			gFBLinePos = 1;
			FBLineInputHash( fileID, (StringPtr)&gFBLineBuf );
		}
	}
	FBParseInput( varType, target );
}

static int FBXSerialRead_Priv( const int serialPortID )
{
	int bytesRead;
	
	bytesRead = read( gFBOpenFiles_A[serialPortID].fileDescriptor, gFBOpenFiles_A[serialPortID].buff + gFBOpenFiles_A[serialPortID].bytesInBuff, gFBOpenFiles_A[serialPortID].buffSize - gFBOpenFiles_A[serialPortID].bytesInBuff );
	gFBOpenFiles_A[serialPortID].bytesInBuff += bytesRead;
	return bytesRead;
}



void FBReadString( const int fileID, void * dest, ByteCount numBytesRequested )
{
	// read #fileID, s; numBytes
	OSStatus	err;
	ByteCount	numBytesRead;
	
	*(char*)dest = 0;
	if ( (numBytesRequested == 0) && (FBlof( fileID, 1 ) > 0) ) numBytesRequested = 1; // read #fileID, s; 0  used to poll serial port

	if ( numBytesRequested )
	{
		err = FBReadFile_Priv( fileID, dest + 1, numBytesRequested, &numBytesRead );
		*(char*)dest= numBytesRead;
		FBCheckFileError( fileID, err );
	}
}



////////////////////////////////////
//              !!!: Output       //
////////////////////////////////////


void FBWriteFile( const int fileID, void * src, ByteCount numBytes ) __attribute__ ((noinline));
void FBWriteFile( const int fileID, void * src, ByteCount numBytes )
{	
// write file[#] deviceID, address&, numBytes& 
	OSStatus		err = noErr;
	
	if ( fileID > 0 && fileID < kMaxFBFiles ) // disk file
	{
		if ( gFBOpenFiles_A[fileID].isOpen && gFBOpenFiles_A[fileID].forkRefNum )
			err = FSWriteFork( gFBOpenFiles_A[fileID].forkRefNum, fsAtMark + pleaseCacheMask, 0, numBytes, src, nil );		
		else
			err = fnOpnErr;
	}
	
	else if ( fileID < 0 && fileID >= -kFBMaxSerialFiles ) // serial
	{
		int				serialPortID;
		ByteCount		bytesSent;

		serialPortID = kMaxFBFiles - fileID;
		if ( gFBOpenFiles_A[serialPortID].isOpen )
		{
			bytesSent = write( gFBOpenFiles_A[serialPortID].fileDescriptor, src, numBytes );
			if ( bytesSent != numBytes ) err = writErr;
		}
		else
			err = fnOpnErr;
	}
	
	FBCheckFileError( fileID, err );
}



void FBWriteString( const int fileID, char * src, ByteCount varLength ) __attribute__ ((noinline));
void FBWriteString( const int fileID, char * src, ByteCount varLength )
{	
	/*
	 write# stringVar[;length]
	 */
	int		i, writeBytes = varLength;
	char	temp[256];
	
	if ( writeBytes > *(unsigned char*)src ) writeBytes = *(unsigned char*)src;
	FBWriteFile( fileID, src + 1, writeBytes );
	writeBytes = varLength - *(unsigned char*)src;
	if ( writeBytes > 0 )
	{
		if ( writeBytes > 255 ) writeBytes = 255;
		for ( i = 0; i < writeBytes; i++ )
		{
			temp[i] = ' ';
		}
		FBWriteFile( fileID, &temp[0], writeBytes );
	}
}



/*
 write# numeric var    big-endian on disk
 */
void FBWriteSwapShort( short s, const unsigned fileID )
{
	s = CFSwapInt16HostToBig( s );
	FBWriteFile( fileID, &s, 2 );
}

void FBWriteSwapLong( long l, const unsigned fileID )
{
	l = CFSwapInt32HostToBig( l );
	FBWriteFile( fileID, &l, 4 );
}

void FBWriteSwapSingle( float f, const unsigned fileID )
{
	CFSwappedFloat32 fs = CFConvertFloatHostToSwapped( f );
	FBWriteFile( fileID, &fs, sizeof( fs ) );
}

void FBWriteSwapDouble( double d, const unsigned fileID )
{
	CFSwappedFloat64 ds = CFConvertDoubleHostToSwapped( d );
	FBWriteFile( fileID, &ds, sizeof( ds ) );
}


// write fField [#]deviceID, handleVar
void FBWriteField( int fileID, Handle inHandle )
{
	long              size;
	
	size = GetHandleSize( inHandle );
	FBWriteSwapLong( size, fileID );
	FBWriteFile( fileID, *inHandle, size );
} 


/*
 Print #printitem,...
 PrintChar - PrintCR
 */

void FBFilePrintChar( char  theChar )
{
	if ( theChar == 13) 		 
	{
		gFBOpenFiles_A[gFBRoute].linePos = 0;
	}
	else
	{
		gFBOpenFiles_A[gFBRoute].linePos++;
	}
	gFBLastFilePos = gFBOpenFiles_A[gFBRoute].linePos;	
	FBWriteFile( gFBRoute, &theChar, 1 );
} 


SInt64 FBFileGetPos()
{
	return gFBLastFilePos;
}


void FBFilePrintTab()
{
	int             spaceCount;
	int             i;
	
	if ( gFBFileTabMode )
	{
		FBFilePrintChar( gFBFileTabMode );
	}
	else
	{
		spaceCount = gFBTabWidth - (gFBOpenFiles_A[gFBRoute].linePos % gFBTabWidth);
		if ( spaceCount < 100 )
		{
			for ( i = 0; i < spaceCount; i++ ) { FBFilePrintChar( 32 ); }
		}
	}
}


void FBFilePrintCR()
{
	FBFilePrintChar( 13 );
} 


void FBFilePrintString( Str255  s_p )
{
	Str255 s;  PSstrcpy( s, s_p );
	gFBStk -= 1;
	gFBOpenFiles_A[gFBRoute].linePos += s[0];
	gFBLastFilePos = gFBOpenFiles_A[gFBRoute].linePos;
	FBWriteFile( gFBRoute, (void*)&s + 1, s[0] );
} 





////////////////////////////////////
//            !!!: Misc stuff     //
////////////////////////////////////


#if !__LP64__

StringPtr FBFilesLastType( void )
{
	/*
	 lastFileType$ = files$
	 */
	return PSmki( gFBFilesInfo.lastType );
}
#endif  /* !__LP64__ */



void FBSetOpenType( Str15  opentype$_p )
{
	Str255 opentype$;  PSstrcpy( opentype$, opentype$_p );
	gFBStk -= 1;
	UInt8             j;
	
	if ( opentype$[0] >= 4 )
	{
		gFBSaveType = 0;
		j = 1;
		do
		{
			gFBSaveType = gFBSaveType << 8;
			gFBSaveType += opentype$[j];
			j++;
		}
		while ( j <= 4 );
	}
	if ( opentype$[0] >= 8 )
	{
		gFBSaveCreator = 0;
		j = 5;
		do
		{
			gFBSaveCreator = gFBSaveCreator << 8;
			gFBSaveCreator += opentype$[j];
			j++;
		}
		while ( j <= 8 );
	}
}


#if !__LP64__

void FBRename( Str255 oldName, Str255 newName, short vRefNum, long dirID )
{
	gFBStk -= 2;
	FSSpec	spec;
	
	if ( FSMakeFSSpec( vRefNum, dirID, oldName, &spec ) == noErr ) FSpRename( &spec, newName );
}


void FBKill( Str255 path, short vRefNum, long parID )
{
	FSSpec	spec;
	
	if ( FSMakeFSSpec( vRefNum, parID, path, &spec ) == noErr ) FSpDelete( &spec );
}

#endif  /* !__LP64__ */


//////////////////////////////////////////
//  !!!: FSRef and CFURLRef helpers     //
//////////////////////////////////////////


static OSStatus Str255ToHFSUniStr_Priv( ConstStr255Param name, TextEncoding encoding, HFSUniStr255 * oUniName )
{
	OSStatus					err;
	ByteCount					uniLength;
	static TextToUnicodeInfo	sTextToUnicodeInfo;
	
	if ( sTextToUnicodeInfo == 0 )
	{
		err = CreateTextToUnicodeInfoByEncoding( encoding, &sTextToUnicodeInfo );
		if ( err ) return err;
	}
	err = ConvertFromPStringToUnicode( sTextToUnicodeInfo, name, 510, &uniLength, oUniName->unicode );
	oUniName->length = uniLength / sizeof( UniChar );
	return err;
}


OSStatus GetFileFSRef( FSRef * parRef, ConstStr255Param name, FSRef * fileRef )
{
	HFSUniStr255	uniName;
	OSStatus		err;
	
	err = Str255ToHFSUniStr_Priv( name, kTextEncodingMacRoman,  &uniName );
	if ( err ) return err;
	err = FSMakeFSRefUnicode( parRef, uniName.length, &uniName.unicode[0], kTextEncodingUnknown, fileRef );
	return err;						  
}


void CopyNameFromFSRef( FSRef * ref, CFStringRef * string )
{
	HFSUniStr255	uniStr255;
	
	*string = NULL;
	if ( FSGetCatalogInfo( ref, kFSCatInfoNone, NULL, &uniStr255, NULL, NULL ) == noErr )
	{
		*string = CFStringCreateWithCharacters( kCFAllocatorDefault, &uniStr255.unicode[0], uniStr255.length );
	}
}


void GetNameAsPascalStringFromFSRef( FSRef * ref, Str255 * name )
{
	HFSUniStr255	uniStr255;
	CFStringRef		string = NULL;
	
	if ( FSGetCatalogInfo( ref, kFSCatInfoNone, NULL, &uniStr255, NULL, NULL ) == noErr )
	{
		string = CFStringCreateWithCharacters( kCFAllocatorDefault, &uniStr255.unicode[0], uniStr255.length );
		if ( string )
		{
			CFStringGetPascalString( string, *name, 256, kCFStringEncodingMacRoman );
			CFRelease( string );
		}
	}
}


void CopyNameFromCFURLRef( CFURLRef url, CFStringRef * string )
{
	FSRef	ref;
	
	*string = NULL;
	if ( CFURLGetFSRef( url, &ref ) ) CopyNameFromFSRef( &ref, string );
}


void GetNameAsPascalStringFromCFURLRef( CFURLRef url, Str255 * name )
{
	FSRef	ref;	
	
	if ( CFURLGetFSRef( url, &ref ) ) GetNameAsPascalStringFromFSRef( &ref, name );
}

#endif /*  !DECARBONATE */

